import { AST, ESLint, Linter } from 'eslint';
import type { Comment } from 'estree';
/** `results` 内で使われているプラグインの名前のリストを洗い出して返す */
export declare function scanUsedPluginsFromResults(results: ESLint.LintResult[]): string[];
export declare type DisableComment = {
    type: 'Block' | 'Line';
    scope: 'next-line' | 'file';
    ruleIds: string[];
    description?: string;
    range: [number, number];
};
/**
 * コメントを ESLint の disable comment としてパースする。
 * disable comment としてパースできなかった場合は undefined を返す。
 *
 * ## 参考: disable comment の構造
 * /* eslint-disable-next-line rule-a, rule-b, rule-c, rule-d -- I'm the rules.
 *    ^^^^^^^^^^^^^^^^^^^^^^^^ ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ ^^ ^^^^^^^^^^^^^^
 *    |                        |                              |  |
 *    header                   |                              |  |
 *                             ruleList                       |  |
 *                                            descriptionHeader  |
 *                                                               description
 */
export declare function parseDisableComment(comment: Comment): DisableComment | undefined;
/**
 * Convert `DisableComment` to comment text.
 */
export declare function toCommentText({ type, scope, ruleIds, description }: Omit<DisableComment, 'range'>): string;
export declare type InlineConfigComment = {
    description?: string;
    rulesRecord: Partial<Linter.RulesRecord>;
    range: [number, number];
};
/**
 * Convert `InlineConfigComment` to comment text.
 */
export declare function toInlineConfigCommentText({ rulesRecord, description }: Omit<InlineConfigComment, 'range'>): string;
/**
 * Create the results with only messages with the specified rule ids.
 * @param results The lint results.
 * @param ruleIds The rule ids.
 * @returns The results with only messages with the specified rule ids
 */
export declare function filterResultsByRuleId(results: ESLint.LintResult[], ruleIds: (string | null)[]): ESLint.LintResult[];
/**
 * push rule ids to the disable comment and return the new comment node.
 * @param comment The comment node to be modified
 * @param ruleIds The rule ids to be added
 * @returns The new comment node
 */
export declare function pushRuleIdsToDisableComment(comment: DisableComment, ruleIds: string[]): DisableComment;
/**
 * Merge the ruleIds and description of the disable comments.
 * @param a The ruleIds and description of first disable comment
 * @param b The ruleIds and description of second disable comment
 * @returns The ruleIds and description of merged disable comment
 */
export declare function mergeRuleIdsAndDescription(a: {
    ruleIds: string[];
    description?: string;
}, b: {
    ruleIds: string[];
    description?: string;
}): {
    ruleIds: string[];
    description?: string;
};
/**
 * Find shebang from the first line of the file.
 * @param sourceCodeText The source code text of the file.
 * @returns The information of shebang. If the file does not have shebang, return null.
 */
export declare function findShebang(sourceCodeText: string): {
    range: AST.Range;
} | null;
//# sourceMappingURL=eslint.d.ts.map